from git import Repo
from envoy import run as envoy_run
from json import loads
from os.path import join, normpath
from unittest2 import TestCase

from accio.commands.send import send
from accio.utils import (
    delete_tempfolder,
    make_tempdir,
    norm_path,
    make_tempfile)


class SendTest(TestCase):
    def test_send_file(self):
        folder = make_tempdir()
        remote = make_tempdir()
        filename = join(folder, 'test.txt')

        repo = Repo.init(folder)
        Repo.init(remote, bare=True)

        f = open(filename, 'w')
        f.write("1\n2\n7\n5\n6")
        f.close()

        repo.git.add('.')
        repo.git.commit(m="testing")
        repo.git.remote('add', 'origin', normpath(remote))
        repo.git.push('-u', 'origin', 'master')

        fin = make_tempfile("1\n2\n3\n4\n5\n6")

        json = send(
            norm_path(filename),
            2,
            5,
            norm_path(fin),
            True,
            norm_path(folder)
        )

        self.assertNotEqual(json, False)
        payload = loads(json)
        self.assertEqual(payload['lines[]'], [2, 4])

        delete_tempfolder(folder)
        delete_tempfolder(remote)



from os import getcwd
from os.path import join
from pep8 import StyleGuide
from unittest2 import TestCase


class LinterTest(TestCase):
    def test_pep8_conformance(self):
        pep8style = StyleGuide(testsuite=True)
        report = pep8style.init_report()

        pep8style.input_dir(dirname=join(getcwd(), 'accio'))

        self.assertEqual(report.total_errors, 0, "Found code style errors.")



from git import Repo
from envoy import run as envoy_run
from json import loads, dumps
from os.path import join
from unittest2 import TestCase
from accio.commands.translate import translate

from accio.utils import (
    delete_tempfolder,
    make_tempdir,
    norm_path,
    make_tempfile)


class TranslateTest(TestCase):
    def test_translate(self):
        folder = make_tempdir()
        filename = 'test.txt'
        full_filename = join(folder, filename)

        repo = Repo.init(folder)

        f = open(full_filename, 'w')
        f.write("1\n2\n7\n5\n6")
        f.close()

        repo.git.add('.')
        repo.git.commit(m="testing")

        f = open(full_filename, 'w')
        f.write("1\n2\n3\n4\n5\n6")
        f.close()

        repo.git.add('.')
        repo.git.commit(m="testing2")

        data = {
            1: 1,
            2: 4,
            3: 5,
            5: 7,
        }

        expected = {
            "1": 1,
            "2": 4,
            "6": 7,
        }

        json = translate(
            "HEAD~",
            "HEAD",
            norm_path(filename),
            norm_path(folder),
            dumps(data)
        )

        payload = loads(json)
        self.assertEqual(payload, expected)

        delete_tempfolder(folder)



from os import unlink, write, close
from shutil import rmtree
from tempfile import mkstemp, mkdtemp


def make_tempfile(content=None):
    """ Creates a temporary file and returns the path. """
    fd, tmpfile = mkstemp()

    if content:
        write(fd, content)

    close(fd)
    return tmpfile


def delete_tempfile(path):
    """ Deletes a temporary file """
    try:
        unlink(path)
    except:
        pass


def delete_tempfolder(path):
    """Deletes a temporary folder """
    try:
        rmtree(path)
    except:
        pass


def norm_path(path):
    return path.replace('\\', '/')


def make_tempdir():
    """Creates a temporary folder directory and returns the path."""
    return mkdtemp()


class cached_property(object):
    """ Taken from Django

    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    Optional ``name`` argument allows you to make cached properties of other
    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )
    """
    def __init__(self, func, name=None):
        self.func = func
        self.name = name or func.__name__

    def __get__(self, instance, type=None):
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res



from sys import stdout, stderr, stdin
from time import time
import os

from accio.common import repo_url
from accio.network import Connection


def getCoeffsForFile(filename):
    repo_path, server_url, filename = repo_url(filename)

    def getLocalFile(filename):
        return repo_path + os.sep + filename

    if server_url is None:
        return False

    print(filename)

    json = None
    try:
        conn = Connection(server_url)
        json = conn.get("/api/coefficients/" + filename).json()
    except Exception as e:
        stderr.write(str(e))
        return None

    return map(
        lambda x: (getLocalFile(x["destination"]), x["coefficient"]),
        json
    )



from json import dumps, loads
from re import sub
from sys import stderr, stdout, stdin
from os.path import relpath

from accio.scm.diff import create_diff, convert_line_numbers
from accio.scm.git import Git


def translate(old_commit, new_commit, filename, repo_path, inp):
    filename = relpath(filename, repo_path)
    repo_path = relpath(repo_path)

    try:
        git = Git(repo_path)
    except Exception as e:
        stderr.write(str(e))
        return

    # Load the line numbers and the times they have been looked at
    line_counts = loads(sub('\s', '', inp))
    # Put those line numbers in a dictionary so we can translate them to the new diff
    line_counts = dict((int(k), v) for k, v in line_counts.iteritems())

    diff = create_diff(
        old_content=git.get_file_content(filename, old_commit),
        new_content=git.get_file_content(filename, new_commit))
    new_lines = convert_line_numbers(diff, line_counts.keys())

    result = {}

    for i, v in enumerate(line_counts.values()):
        if new_lines[i] is not None:
            result[new_lines[i]] = v

    return dumps(result)



from requests import post, get
from urlparse import urljoin


class Connection(object):
    """ Abstracted requests object to make it easy to test """

    def __init__(self, url):
        self.__url = url

    def post(self, payload, path):
        """ Sends a payload to the connection's URL with a payload """
        return post(urljoin(self.__url, path), data=payload)

    def get(self, path):
        return get(urljoin(self.__url, path))



from envoy import run as envoy_run
from os.path import normpath

from accio.scm import SCMBase


class Git(SCMBase):
    def _get_last_pushed_commit(self):
        r = envoy_run('git rev-list HEAD')
        commits = r.std_out.split('\n')
        for commit in commits:
            r = envoy_run('git branch -r --contains %s' % commit)
            if r.std_out != "":
                return commit

        return None

    def _get_file_content(self, file_path, commit):
        r = envoy_run('git show %s:"%s"' % (
            commit,
            normpath(file_path)))

        return r.std_out

    def _get_name(self):
        r = envoy_run('git config user.name')
        return r.std_out.strip()

    def _get_email(self):
        r = envoy_run('git config user.email')
        return r.std_out.strip()

    def _get_branch(self):
        r = envoy_run('git rev-parse --abbrev-ref HEAD')
        return r.std_out.strip()



from json import dumps
from sys import stdout, stderr, stdin
from time import time
from os.path import relpath

from accio.common import repo_url
from accio.network import Connection
from accio.scm.diff import create_diff, convert_line_numbers
from accio.scm.git import Git


def send(filename, start, end, buffer, debug=False, use_repo_path=None):
    repo_path, server_url, new_filename = repo_url(filename)

    if use_repo_path:
        # we want to override the repo_path for some reason (testing usually)
        repo_path = use_repo_path

    if new_filename:
        filename = new_filename

    if use_repo_path:
        # if we have overriden it, our filepath is no longer relative
        filename = relpath(filename, repo_path)

    if server_url is None and not debug:
        return False

    try:
        git = Git(repo_path)
    except Exception as e:
        stderr.write(str(e))
        return None

    commit = git.get_last_pushed_commit()

    f = open(buffer, "r")
    wip = f.read()

    lines = convert_line_numbers(
        create_diff(
            old_content=wip,
            new_content=git.get_file_content(filename, commit)),
        lines=range(start, end + 1))
    lines = [x for x in lines if x is not None]

    payload = {
        "timestamp": int(time() * 1000),
        "file": filename,
        "name": git.name,
        "email": git.email,
        "branch": git.branch,
        "commit": commit,
        "lines[]": lines,
    }

    if not debug:
        try:
            conn = Connection(server_url)
            conn.post(path="/api/snapshot", payload=payload)
        except Exception as e:
            stderr.write(str(e))
            return False
    return dumps(payload)



from accio.parser import build_parser
from accio.commands.translate import translate
from accio.commands.send import send
from accio.commands.pull import getCoeffsForFile
from sys import stderr, stdout, stdin, argv



import xmlrpclib
from SimpleXMLRPCServer import SimpleXMLRPCServer

OPTIONS = {
    'command': str,
    '--old_commit': str,
    '--new_commit': str,
    '--filename': str,
    '--repo_path': str,
}

def main():
    # We want our accio to be able to just use the translate function. 
    # This is used on the backend server to translate diffs between commits
    if len(argv) > 1:
        parser = build_parser(OPTIONS)
        (opt, args) = parser.parse_known_args()
        if opt.command != "translate":
            print "Invalid command, either run the server with just 'accio' or provide the correct parameters for 'translate'"
            return 1
        # Pass our paremeters down and read the map of line numbers -> number of times looked at from stdin
        print translate(opt.old_commit, opt.new_commit, opt.filename, opt.repo_path, stdin.read())
    else:
        # Otherwise we are running the accio on the user's machine so set it up to send the info to our server
        server = SimpleXMLRPCServer(("localhost", 7432))
        server.register_function(send, "snapshot")
        server.register_function(translate, "translate")
        server.register_function(getCoeffsForFile, "coefficients")

        server.serve_forever()

    return 0



from abc import ABCMeta, abstractmethod
from os import chdir, getcwd
from os.path import relpath


class SCMBase(object):
    __metaclass__ = ABCMeta

    def __init__(self, repo_path):
        self.__repo_path = repo_path

        original_path = getcwd()
        chdir(self.__repo_path)

        self.name = self._get_name()
        self.email = self._get_email()
        self.branch = self._get_branch()

        chdir(original_path)

        if self.email == "":
            raise Exception('Snapshot not sent: No email has been provided')

    def relative_file_path(self, file_path):
        """ Returns the filepath relative to the repository """
        return relpath(file_path, self.__repo_path)

    @abstractmethod
    def _get_name(self):
        """ Gets the name associated with the SCM """
        return

    @abstractmethod
    def _get_email(self):
        """ Gets the email associated with the SCM """
        return

    @abstractmethod
    def _get_branch(self):
        """ Gets the current branch of the SCM """
        return

    def get_last_pushed_commit(self):
        """
        Retrieve commit id of the most recent commit pushed to remote from
        the current directory.
        """
        original_path = getcwd()
        chdir(self.__repo_path)
        result = self._get_last_pushed_commit()
        chdir(original_path)

        return result

    @abstractmethod
    def _get_last_pushed_commit(self):
        return

    def get_file_content(self, file_path, commit):
        """
        Returns the path to a temporary file containing the contents of the
        specified file at the base commit. The same path should be returned
        every time.
        """
        original_path = getcwd()
        chdir(self.__repo_path)
        result = self._get_file_content(file_path, commit)
        chdir(original_path)

        return result

    @abstractmethod
    def _get_file_content(self, file_path, commit):
        return



from argparse import ArgumentParser

from accio import ACCIO_MAIN


def build_parser(options):
    parser = ArgumentParser(
        prog=ACCIO_MAIN,
        usage='%(prog)s <command> [<args>]',
        add_help=False)

    for k, v in options.iteritems():
        if (v == bool):
            parser.add_argument(k, action="store_true")
        else:
            parser.add_argument(k, type=v)

    return parser



from difflib import ndiff

# we might want to integrate this for patience diffing
# http://bazaar.launchpad.net/~bzr-pqm/bzr/bzr.dev/view/head:/bzrlib/_patiencediff_py.py


def create_diff(new_content, old_content):
    """ Returns the diff between the new and old file """
    diff = ndiff(
        a=old_content.splitlines(),
        b=new_content.splitlines(),
        linejunk=None,
        charjunk=None)

    delta = []
    for line in diff:
        if line.startswith('  '):
            delta.append('-')
        elif line.startswith('+ '):
            delta.append('>')
        elif line.startswith('- '):
            delta.append('<')

    return delta


def add_to_result(lines, result, old_line, new_line):
    if lines[0] == old_line:
        result.append(new_line)
        lines.pop(0)


# get line numbers in second arg to diff, given a range in first arg
def convert_line_numbers(diff, lines):
    lines.sort()

    left_line = 0
    right_line = 0
    result = []
    for line in diff:
        if line == '<':
            left_line += 1
            add_to_result(lines, result, left_line, None)
        elif line == '>':
            right_line += 1
        else:
            left_line += 1
            right_line += 1
            add_to_result(lines, result, left_line, right_line)

        if not len(lines):
            break

    # Ensure that our array sizes are the same if we ran out of lines in diff
    # without matching all of the lines
    for i in range(len(lines)):
        result.append(None)

    return result



import os
from os import path


def walk_up(bottom):
    """
    mimic os.walk, but walk 'up'
    instead of down the directory tree
    from: https://gist.github.com/zdavkeos/1098474
    """

    bottom = path.realpath(bottom)

    # get files in current dir
    try:
        names = os.listdir(bottom)
    except Exception as e:
        print e
        return

    dirs, nondirs = [], []
    for name in names:
        if path.isdir(path.join(bottom, name)):
            dirs.append(name)
        else:
            nondirs.append(name)

    yield bottom, dirs, nondirs

    new_path = path.realpath(path.join(bottom, '..'))

    # see if we are att the top
    if new_path == bottom:
        return

    for x in walk_up(new_path):
        yield x


def repo_url(file):
    """
    Returns the list of server urls associated with a file path. If the path
    is not in a supported repository, None is returned.
    """
    for dircontents in walk_up(path.dirname(file)):
        for filename in dircontents[2]:
            dirname = dircontents[0]
            if filename == ".take2rc":
                with open(dirname + os.sep + filename) as f:
                    return dirname, f.read(), path.relpath(file, dirname)
    return None, None, None



from unittest2 import TestCase

from accio.scm.diff import create_diff, convert_line_numbers


class DiffTest(TestCase):
    def test_diff_additions(self):
        self.assertEqual(convert_line_numbers(
            ['-', '-', '>', '>', '-', '-'],
            range(1, 4)),
            [1, 2, 5])

    def test_diff_subtractions(self):
        self.assertEqual(convert_line_numbers(
            ['-', '-', '<', '-', '<', '-', '-'],
            range(2, 7)),
            [2, None, 3, None, 4])

    def test_diff_modifications(self):
        self.assertEqual(convert_line_numbers(
            ['-', '-', '<', '>', '-', '-'],
            range(1, 6)),
            [1, 2, None, 4, 5])

        self.assertEqual(convert_line_numbers(
            ['-', '-', '<', '<', '>', '-', '-'],
            range(2, 6)),
            [2, None, None, 4])

        self.assertEqual(convert_line_numbers(
            ['-', '-', '<', '<', '>', '>', '>', '-', '-', '-', '-'],
            range(5, 9)),
            [6, 7, 8, 9])

    def test_create_diff(self):
        self.assertEqual(
            create_diff(
                new_content="a\nb\nc\n",
                old_content="a\nc\n"),
            ['-', '>', '-'])

        self.assertEqual(
            create_diff(
                new_content="a\nc\n",
                old_content="a\nb\nc\n"),
            ['-', '<', '-'])



from os import chdir
from os.path import dirname

try:
    from setuptools import setup, find_packages
except ImportError:
    # setuptools wasn't available, so install and try again
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

from accio import get_package_version, ACCIO_MAIN

root_dir = dirname(__file__)

if root_dir != "":
    chdir(root_dir)

setup(
    name='accio',
    version=get_package_version(),
    author='Matt Maclean',
    author_email='matthewcmaclean@gmail.com',
    description='Sends information from code editors to server',
    entry_points={
        'console_scripts': {
            ACCIO_MAIN + ' = accio.cli:main',
        }
    },
    url='https://github.com/isovector/take2', # use the URL to the github repo
    packages=find_packages(exclude="test"),
    include_package_data=True,
    install_requires=[
        'Envoy>=0.0.2',
        'requests>=2.3.0',
    ],
    test_suite='unittest2.collector',
    tests_require=[
        'unittest2>=0.5.1',
        'pep8>=1.5.7',
        'gitpython>=0.3.1',
    ],
)
